# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import inspect
from typing import Callable, Optional, Any, List, Tuple, TYPE_CHECKING
if TYPE_CHECKING: # Avoid circular import when a type checker analyses the code
    from ..electronics._camera import CherenkovSiPMCamera

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from ._iactsim_style import iactsim_style

@iactsim_style
def plot_sipm_modules(
    camera : 'CherenkovSiPMCamera',
    plotf : Callable[[int, Optional[matplotlib.axes.Axes]], Any],
    figsize: float = 10,
    sep_factor: float = 1.25,
    skip: Optional[List[int]] = None
    ) -> Tuple[matplotlib.figure.Figure, List[matplotlib.axes.Axes]]:
    """
    Creates a custom plot for each module based on its ID using a provided plotting function.

    Parameters
    ----------
    camera : CherenkovSiPMCamera
        Camera from which extract geometry info. 
    plotf : Callable[[int, Optional[matplotlib.axes.Axes]], Any]
        A callable that takes the module index (an integer starting from 0)
        as its first argument, and optionally a matplotlib Axes
        object as the second argument.  This function is responsible for
        generating the plot content for each module.  If `plotf` only accepts
        one argument, it's called with the module index only.  If it
        accepts two, it's called with the module index and the
        corresponding Axes object.
    figsize : float
        The size of the figure (both width and height) in inches. By default 10
    sep_factor : float
        A scaling factor that controls the separation between
        modules in the plot.  A larger `sep_factor` results in more
        space between modules. By default 1.25.
    skip : Optional[List[int]], default=None
        An optional list of module indices to skip plotting.

    Returns
    -------
    Tuple[matplotlib.figure.Figure, List[matplotlib.axes.Axes]]
        A tuple containing:
            - fig: The Matplotlib figure object.
            - axes: A list of Matplotlib Axes objects, one for each plotted module.

    Raises
    ------
    RuntimeError
        If `camera.geometry` is None (no camera geometry is defined).
    """
    if camera.geometry is None:
        raise(RuntimeError("A camera geometry must be defined in order to show SiPM module plots."))

    if skip is None:
        skip = []

    # Scale real coordinates into figure coordinates
    x, y,_ = camera.geometry.modules_p.T
    side = camera.geometry.module_side

    scale = (max(np.sqrt(x**2+y**2))*2+side*np.sqrt(2))
    x = (x-min(x)) / scale
    y = (y-min(y)) / scale
    
    side = side / scale / sep_factor
    h_side = 0.5*side

    fig = plt.figure(figsize=(figsize, figsize))
    axes = []
    for module_id in range(len(x)):
        if module_id in skip:
            continue
        ax = fig.add_axes([x[module_id]+h_side, y[module_id]+h_side, side, side])

        n_parameters = len(inspect.signature(plotf).parameters)
        if n_parameters == 1:
            plt.sca(ax)
            plotf(module_id)
        else:
            plotf(module_id, ax)
        
        axes.append(ax)
    
    return fig, axes

@iactsim_style
def plot_sipm_module_pixels(
        camera : 'CherenkovSiPMCamera',
        plotf : Callable[[int, Optional[matplotlib.axes.Axes]], Any], 
        subplot_size : float = 4
        ) -> Tuple[matplotlib.figure.Figure, List[matplotlib.axes.Axes]]:

    """
    Creates a custom plot for each pixel of a module based on its ID using a provided plotting function.

    Parameters
    ----------
    camera : CherenkovSiPMCamera
        Camera from which extract geometry info. 
    plotf : Callable[[int, Optional[matplotlib.axes.Axes]], Any]
        A callable that takes the pixel index (an integer starting from 0)
        as its first argument, and optionally a matplotlib Axes
        object as the second argument.  This function is responsible for
        generating the plot content for each pixel.  If `plotf` only accepts
        one argument, it's called with the pixel index only.  If it
        accepts two, it's called with the pixel index and the
        corresponding Axes object.
    subplot_size : float
        The size of the pixel subplot (both width and height) in inches. By default 4

    Returns
    -------
    Tuple[matplotlib.figure.Figure, List[matplotlib.axes.Axes]]
        A tuple containing:
            - fig: The Matplotlib figure object.
            - axes: A list of Matplotlib Axes objects, one for each plotted pixel.

    Raises
    ------
    RuntimeError
        If `camera.geometry` is None (no camera geometry is defined).
    """

    if camera.geometry is None:
        raise(RuntimeError("A camera geometry must be defined in order to show SiPM module plots."))

    pix_per_side = round(camera.geometry.module_side/(camera.geometry.pixel_active_side+camera.geometry.pixels_separation))

    fig_size = pix_per_side * subplot_size
    fig, axes = plt.subplots(pix_per_side,pix_per_side,figsize=(fig_size,fig_size))
    axes = [ax for axs in axes[::-1] for ax in axs] # flatten

    n_parameters = len(inspect.signature(plotf).parameters)

    for pixel_id,ax in enumerate(axes):        
        if n_parameters == 1:
            plt.sca(ax)
            plotf(pixel_id)
        else:
            plotf(pixel_id, ax)
    
    return fig, axes


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import numpy as np
import vtk

from ._vtk_utils import (
    create_aspheric_surface_actor,
    create_cylindric_surface_actor,
    quit
)

from ..optics._surface_misc import SurfaceType, SurfaceShape
from ..optics._surface import SurfaceVisualProperties

from ..optics._cpu_transforms import local_to_telescope_rotation, local_to_pointing_rotation

from .._iact import IACT

from ..optics._optical_system import OpticalSystem


class VTKOpticalSystem():
    """Class to viasualize the geometry of an optical system.
    Each surface actor can be accessed from :py:attr:`actors` attribute after a :py:meth:`update` call.

    Parameters
    ----------
    optical_system : OpticalSystem or IACT
        Optical sistem for which visualize surfaces.
    resolution : float, optional
            Objects mesh resolution (in mm). By default 10 mm.

    Notes
    -----
    If you perform some operation on an actor, make sure to call :py:meth:`start_render` with ``update=False``, otherwise the actors will be replaced.
    
    """
    def __init__(self, optical_system, resolution=None):
        self.actors = {}
        """Dictionary of surface name: surface actor."""

        if issubclass(type(optical_system), OpticalSystem):
            self.os = optical_system
            self._tel_to_local_rot =  np.eye(3)
            self._translation_vector = np.zeros(3)
        elif issubclass(type(optical_system), IACT):
            self.os = optical_system.optical_system
            self._tel_to_local_rot =  optical_system.local_to_telescope_rotation.T
            self._translation_vector = optical_system.position
        else:
            raise(ValueError("optical system must be an instance of OpticalSystem or IACT."))

        """Optical system to be visualized."""
        
        self._mirror_props = SurfaceVisualProperties(
            color=(0.6, 0.6, 0.6),
            opacity=1.0,
            specular=0.4,
            wireframe=False,
            resolution=None,
            visible=True
        )

        self._opaque_props = SurfaceVisualProperties(
            color=(0.4, 0.4, 0.4),
            opacity=1.0,
            specular=0.1,
            wireframe=False,
            resolution=None,
            visible=True
        )

        self._sensitive_props = SurfaceVisualProperties(
            color=(0.55, 0.55, 0.55),
            opacity=1.0,
            specular=0.2,
            wireframe=False,
            resolution=None,
            visible=True
        )

        self._refractive_props = SurfaceVisualProperties(
            color=(0.5, 0.5, 1.0),
            opacity=0.2,
            specular=0.3,
            wireframe=False,
            resolution=None,
            visible=True
        )
                
        """Dictionary to custumize surface colors based on surface type."""

        # Wireframe
        self.wireframe = False
        """Whether to use wireframe representation by default."""

        # Window size
        self.window_size = (1024,1024)
        """Window size in pixel."""

        self._resolution = resolution # mm
        """Objects mesh resolution (in mm)."""

        self._ray_opacity = None
        """Ray opacity."""

        self._update()

        self._apply_global_transform()
    
    def _update(self):
        """Generate all surface actors.
        """
        self.actors = {}
        for i,s in enumerate(self.os):
            vp = s.visual_properties

            res = self._resolution
            if vp.resolution is not None:
                res = vp.resolution
            
            if s._shape == SurfaceShape.CYLINDRICAL:
                actor = create_cylindric_surface_actor(s, res)
            else:
                actor = create_aspheric_surface_actor(s, res)

            transform = vtk.vtkTransform()
            transform.Translate(*s.position)
            R = s.get_rotation_matrix()
            vtk_rotation_matrix = vtk.vtkMatrix4x4()
            for i in range(3):
                for j in range(3):
                    vtk_rotation_matrix.SetElement(i, j, R[j, i])
            transform.PreMultiply()
            transform.Concatenate(vtk_rotation_matrix)
            transform.Update()
            actor.SetUserTransform(transform)

            prop = actor.GetProperty()

            is_sensitive = s.type in [
                SurfaceType.SENSITIVE,
                SurfaceType.SENSITIVE_BACK,
                SurfaceType.SENSITIVE_FRONT,
                SurfaceType.REFLECTIVE_SENSITIVE,
                SurfaceType.REFLECTIVE_SENSITIVE_BACK,
                SurfaceType.REFLECTIVE_SENSITIVE_FRONT
            ]
            dvp = None
            if s.type in [SurfaceType.REFLECTIVE, SurfaceType.REFLECTIVE_BACK, SurfaceType.REFLECTIVE_FRONT]:
                dvp = self._mirror_props
            elif s.type == SurfaceType.OPAQUE:
                dvp = self._opaque_props
            elif is_sensitive:
                dvp = self._sensitive_props
            elif s.type == SurfaceType.REFRACTIVE:
                dvp = self._refractive_props
            
            if self.wireframe or vp.wireframe:
                prop.SetRepresentationToWireframe()

            if vp.visible is not None:
                actor.SetVisibility(vp.visible)

            if vp.color is not None:
                prop.SetColor(*vp.color)
            else:
                prop.SetColor(*dvp.color)
            
            if vp.opacity is not None:
                prop.SetOpacity(vp.opacity)
            else:
                prop.SetOpacity(dvp.opacity)

            if vp.specular is not None:
                prop.SetSpecular(vp.specular)
            else:
                prop.SetSpecular(dvp.specular)

            prop.SetAmbient(0.3)
            prop.SetDiffuse(0.5)
            prop.SetSpecularPower(10)

            self.actors[s.name] = actor

    @staticmethod
    def _get_sigmoid_opacity(n_rays, n_50=1000, k=1.5, min_opacity=0.005):
        """
        Calculates ray opacity using a descending sigmoid curve.
        
        Parameters
        ----------
        n_rays : int
            Number of rays being plotted.
        n_50 : int
            The 'knee' of the curve: number of rays where opacity is ~0.5.
        k : float
            Steepness of the curve (the greater the steeper the curve). 
        min_opacity : float
            Minimum visibility floor.
        """
        if n_rays <= 0: return 0.0
        
        # Calculate sigmoid
        opacity = min_opacity + (1.0 - min_opacity) / (1 + (n_rays / n_50)**k)
        
        return opacity

    def start_render(self, camera_position=None, focal_point=None, view_up=(0, 0, 1), orthographic=False):
        """Render the optical system geometry on a VTK window.

        Parameters
        ----------
        camera_position : tuple or list, optional
            (x, y, z) coordinates for the camera position.
        focal_point : tuple, list, or str, optional
            If tuple/list: (x, y, z) coordinates to look at.
            If str: The name of the surface in self.os to look at.
            If None: Defaults to center of system.
        view_up : tuple or list, optional
            (x, y, z) vector defining the "up" direction. Default is (0, 0, 1).
        orthographic : bool, optional
            If True, start with a parallel projection. 
            If False, start with a perspective projection. Default is False.
        """

        # Set opacity based on number of ray actors
        # N.B.: this overestimates the number of photons.
        # N.B.: _set_ray_opacity will not update _ray_opacity
        #       at each start render the opacity will be set automatically
        #       unless an opacity is specified with set_ray_opacity
        if self._ray_opacity is None:
            n_rays = 0
            for actor in self.actors:
                if actor.startswith('rays'):
                    n_rays += self.actors[actor].GetMapper().GetDataSetInput().GetLines().GetNumberOfCells()
            opacity = self._get_sigmoid_opacity(n_rays)
            self._set_ray_opacity(opacity)

        # Rendering
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)

        inf = float('inf')
        sys_bounds = [inf, -inf, inf, -inf, inf, -inf]
        has_system_actors = False

        for name, actor in self.actors.items():
            renderer.AddActor(actor)
            
            # Include this actor in the bounds calculation
            if not name.startswith("rays"):
                has_system_actors = True
                b = actor.GetBounds() # (xmin, xmax, ymin, ymax, zmin, zmax)
                sys_bounds[0] = min(sys_bounds[0], b[0])
                sys_bounds[1] = max(sys_bounds[1], b[1])
                sys_bounds[2] = min(sys_bounds[2], b[2])
                sys_bounds[3] = max(sys_bounds[3], b[3])
                sys_bounds[4] = min(sys_bounds[4], b[4])
                sys_bounds[5] = max(sys_bounds[5], b[5])

        # Display instructions
        desc = [
            "Press: 'e/q' to exit/quit",
            "Press: 'w' or 's' to switch to wireframe or surface representation",
            "Press: 'r' to reset the camera zoom",
            "Press: 'o' to toggle orthographic/perspective view",
            "Press: 'p' to pick/unpick actor under cursor",
            "Press: 'f' fly to the picked actor",
            "Press: 'h' to hide the picked actors",
            "Press: 'P' to print the current view",
            "Press: 'i' to isolate picked actors (and hide others)",
            "Press: 'a' to toggle local axes for the single selected actor", # New instruction
            "Press: 'Esc' to clear selection and actions",
            "Press: 'x', 'y' or 'z' to align the up-vector to the desired axis",
        ]
        desc_has_gpu_info = False
        textActor = vtk.vtkTextActor()
        textActor.SetInput('\n'.join(desc))
        position_coordinate = textActor.GetPositionCoordinate()
        position_coordinate.SetCoordinateSystemToNormalizedViewport()
        position_coordinate.SetValue(0.01, 0.99)
        textActor.GetTextProperty().SetJustificationToLeft()
        textActor.GetTextProperty().SetVerticalJustificationToTop()
        textActor.GetTextProperty().SetFontSize(18)
        textActor.GetTextProperty().SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        textActor.GetTextProperty().SetFontFamily(vtk.VTK_COURIER)
        textActor.SetVisibility(False)
        renderer.AddActor2D(textActor)

        hintActor = vtk.vtkTextActor()
        hintActor.SetInput("Press '?' for instructions")
        hprop = hintActor.GetTextProperty()
        hprop.SetFontSize(16)
        hprop.SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        hprop.SetFontFamily(vtk.VTK_COURIER)
        hprop.SetJustificationToLeft()
        hprop.SetVerticalJustificationToTop()
        hcoord = hintActor.GetPositionCoordinate()
        hcoord.SetCoordinateSystemToNormalizedViewport()
        hcoord.SetValue(0.01, 0.99)
        hintActor.SetVisibility(True)
        renderer.AddActor2D(hintActor)

        render_window = vtk.vtkRenderWindow()
        render_window.SetWindowName(self.os.name)
        render_window.AddRenderer(renderer)
        render_window.SetSize(*self.window_size)
        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)

        # Camera
        cam_style = vtk.vtkInteractorStyleTrackballCamera()
        render_window_interactor.SetInteractorStyle(cam_style)

        # Axes
        axes = vtk.vtkAxesActor()
        widget = vtk.vtkOrientationMarkerWidget()
        rgba = [0] * 4
        vtk.vtkNamedColors().GetColor('Carrot', rgba)
        widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
        widget.SetOrientationMarker(axes)
        widget.SetInteractor(render_window_interactor)
        widget.SetEnabled(1)
        widget.InteractiveOn()

        # Surface sxes widget (hidden by default)
        local_widget = vtk.vtkOrientationMarkerWidget()
        local_widget.SetOutlineColor(0.0, 1.0, 0.0) # Green outline
        local_widget.SetInteractor(render_window_interactor)
        local_widget.SetViewport(0.8, 0.0, 1.0, 0.2) # Bottom Right
        local_widget.SetEnabled(0)
        local_widget.InteractiveOn()

        selected_actors = {} # Store picked actors and their original color
        is_isolated = False
        picker = vtk.vtkCellPicker()
        picker.SetTolerance(0.005)

        def perform_view_reset(up_vector):
            active_camera = renderer.GetActiveCamera()
            
            if not active_camera:
                return

            # Get current view plane normal
            vpn = active_camera.GetViewPlaneNormal()
            
            # Compute dot product
            dot = sum(vpn[i] * up_vector[i] for i in range(3))
            
            # Parallel vectors
            if abs(dot) > 0.95:
                fp = active_camera.GetFocalPoint()
                dist = active_camera.GetDistance()
                
                # Move camera to a perpendicular axis based on target UP
                if abs(up_vector[2]) > 0.9: 
                    active_camera.SetPosition(fp[0], fp[1] - dist, fp[2])
                elif abs(up_vector[1]) > 0.9:
                    active_camera.SetPosition(fp[0] - dist, fp[1], fp[2])
                else:
                    active_camera.SetPosition(fp[0], fp[1], fp[2] + dist)

                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()

            active_camera.SetViewUp(*up_vector)
            active_camera.OrthogonalizeViewUp() 
            render_window.Render()

        def key_press_callback(caller, event):
            nonlocal desc_has_gpu_info
            nonlocal is_isolated

            interactor = caller
            key_sym = interactor.GetKeySym()
            
            if key_sym is None:
                return
            
            # Reset view ('r')
            # Override deafult action to not include rays
            if key_sym.lower() == 'r':
                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()
                render_window.Render()
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")
            
            # Instructions ('?')
            elif key_sym == 'question':
                if not desc_has_gpu_info:
                    # Fetch GPU Info
                    info = render_window.ReportCapabilities()
                    gpu_lines = []
                    for line in info.splitlines():
                        if "OpenGL vendor" in line or "OpenGL renderer" in line:
                            gpu_lines.append(line)
                    # Update the TextActor
                    full_text = desc + gpu_lines
                    textActor.SetInput('\n'.join(full_text))
                desc_has_gpu_info = True

                current_vis = textActor.GetVisibility()
                # If currently visible, make invisible.
                textActor.SetVisibility(not current_vis)
                # Hint is always the opposite of Instructions
                hintActor.SetVisibility(current_vis)
                render_window.Render()

            # Orthographic/perspective ('o')
            elif key_sym.lower() == 'o':
                cam = renderer.GetActiveCamera()
                if cam.GetParallelProjection():
                    # Switching from parallel to perspective
                    scale = cam.GetParallelScale()
                    fov = cam.GetViewAngle()
                    if fov > 0:
                        target_dist = scale / np.tan(np.deg2rad(fov) / 2.0)
                        fp = np.array(cam.GetFocalPoint())
                        pos = np.array(cam.GetPosition())
                        direction = fp - pos
                        dist_mag = np.linalg.norm(direction)
                        if dist_mag > 0:
                            direction_norm = direction / dist_mag
                            cam.SetPosition(fp - direction_norm * target_dist)
                    cam.SetParallelProjection(False)
                else:
                    # Switching from perspective to Parallel
                    dist = cam.GetDistance()
                    fov = cam.GetViewAngle()
                    new_scale = dist * np.tan(np.deg2rad(fov) / 2.0)
                    cam.SetParallelScale(new_scale)
                    cam.SetParallelProjection(True)
                
                renderer.ResetCameraClippingRange()
                render_window.Render()
            
            # Selection ('p')
            # Override default action to pick more than one actor 
            elif key_sym == 'p':
                click_pos = interactor.GetEventPosition()
                picker.Pick(click_pos[0], click_pos[1], 0, renderer)
                picked_actor = picker.GetActor()
                if picked_actor:
                    if picked_actor in selected_actors:
                        # Deselect
                        original_color = selected_actors.pop(picked_actor)
                        picked_actor.GetProperty().SetColor(*original_color)
                        
                        # Disable local axes if deselecting
                        if local_widget.GetEnabled():
                            local_widget.SetEnabled(0)
                    else:
                        # Select
                        current_color = picked_actor.GetProperty().GetColor()
                        selected_actors[picked_actor] = current_color
                        picked_actor.GetProperty().SetColor(0.0, 1.0, 0.0) # Green
                        
                        # Disable local axes if picking new/multiple
                        if local_widget.GetEnabled():
                            local_widget.SetEnabled(0)
                render_window.Render()
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")
            
            # Toggle surface axes ('a')
            elif key_sym.lower() == 'a':
                if local_widget.GetEnabled():
                    local_widget.SetEnabled(0)
                elif len(selected_actors) == 1:
                    actor = list(selected_actors.keys())[0]
                    
                    # Create local axes actor
                    local_axes = vtk.vtkAxesActor()
                    
                    # Get the actor full transform
                    full_matrix = actor.GetMatrix()
                    
                    rotation_matrix = vtk.vtkMatrix4x4()
                    rotation_matrix.DeepCopy(full_matrix)
                    rotation_matrix.SetElement(0, 3, 0.0)
                    rotation_matrix.SetElement(1, 3, 0.0)
                    rotation_matrix.SetElement(2, 3, 0.0)
                    
                    # Apply rotation to the axes actor
                    local_axes.SetUserMatrix(rotation_matrix)
                    
                    # Update widget
                    local_widget.SetOrientationMarker(local_axes)
                    local_widget.SetEnabled(1)
                    
                    render_window.Render()
                else:
                     print("Select exactly one surface (press 'p') to visualize its local axes.")
                
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")

            # Fly-to ('f')
            # Override default action to take into account more than one picked actor
            elif key_sym.lower() == 'f':
                if selected_actors:
                    # Calculate center of all selected actors
                    sel_bounds = [inf, -inf, inf, -inf, inf, -inf]
                    for actor in selected_actors:
                        b = actor.GetBounds()
                        sel_bounds[0] = min(sel_bounds[0], b[0])
                        sel_bounds[1] = max(sel_bounds[1], b[1])
                        sel_bounds[2] = min(sel_bounds[2], b[2])
                        sel_bounds[3] = max(sel_bounds[3], b[3])
                        sel_bounds[4] = min(sel_bounds[4], b[4])
                        sel_bounds[5] = max(sel_bounds[5], b[5])
                    center_x = 0.5 * (sel_bounds[0] + sel_bounds[1])
                    center_y = 0.5 * (sel_bounds[2] + sel_bounds[3])
                    center_z = 0.5 * (sel_bounds[4] + sel_bounds[5])
                    
                    interactor.FlyTo(
                        renderer,
                        center_x, center_y, center_z
                    )

                    render_window.Render()
                    interactor.SetKeySym("") 
                    interactor.SetKeyCode("\0")

            # Hide picked actors ('h')
            elif key_sym.lower() == 'h':
                # Create a reverse lookup to find surface names from actor objects
                actor_to_name = {v: k for k, v in self.actors.items()}

                # Disable local axes
                local_widget.SetEnabled(0)

                for actor in list(selected_actors.keys()):
                    # Hide the surface (reset the original color before removing from picked list)
                    original_color = selected_actors.pop(actor)
                    actor.GetProperty().SetColor(*original_color)
                    actor.SetVisibility(False)
                    
                    # Find and hide associated hits
                    if actor in actor_to_name:
                        surface_name = actor_to_name[actor]
                        # Look for keys starting with "hits_" + surface_name
                        target_prefix = f"hits_{surface_name}"
                        for key, hit_actor in self.actors.items():
                            if key == target_prefix or key.startswith(target_prefix + "_"):
                                hit_actor.SetVisibility(False)

                render_window.Render()

            # Isolate selected ('i')
            elif key_sym.lower() == 'i':
                # Create a reverse lookup to find surface names from actor objects
                actor_to_name = {v: k for k, v in self.actors.items()}

                # Disable local axes
                local_widget.SetEnabled(0)

                # If we have a selection, isolate it and then deselect
                if selected_actors:
                    is_isolated = True
                    # Hide everything not in selection
                    for actor_name, actor in self.actors.items():
                        # Always keep rays visible
                        if actor_name.startswith('rays'):
                            actor.SetVisibility(True)
                            continue

                        for sel_actor in selected_actors:
                            hit_prefix = f"hits_{actor_to_name[sel_actor]}"
                            # Keep the surface
                            if sel_actor == actor:
                                actor.SetVisibility(True)
                                # Restore color
                                orig_col = selected_actors[actor]
                                actor.GetProperty().SetColor(*orig_col)
                            # Keep surface hits
                            elif actor_name == hit_prefix or actor_name.startswith(hit_prefix+'_'):
                                actor.SetVisibility(True)
                            # Hide others
                            else:
                                actor.SetVisibility(False)
                    
                    # Clear selection since we have now no active targets
                    selected_actors.clear()
                    
                # If nothing selected, 'i' acts as "Show All" if currently isolated
                elif is_isolated:
                    is_isolated = False
                    for actor in self.actors.values():
                        actor.SetVisibility(True)
                
                render_window.Render()

                # Consume the key, otherwise the axes will disappear (IDKW)
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")

            # Clear selections ('Escape')
            elif key_sym == 'Escape':
                is_isolated = False
                
                # Restore colors of selected
                for actor, color in selected_actors.items():
                    actor.GetProperty().SetColor(*color)
                selected_actors.clear()
                
                # Restore visibility based on visual_properties
                surface_lookup = {s.name: s for s in self.os}

                for name, actor in self.actors.items():
                    should_be_visible = True
                    
                    # If this actor corresponds to a defined surface, check its properties
                    if name in surface_lookup:
                        vp = surface_lookup[name].visual_properties
                        if vp.visible is False:
                            should_be_visible = False
                    
                    actor.SetVisibility(should_be_visible)
                
                # Disable local axes
                local_widget.SetEnabled(0)
                
                render_window.Render()

            # Axes and modes
            elif key_sym.lower() == 'y':
                perform_view_reset((0,1,0))
            elif key_sym.lower() == 'x':
                perform_view_reset((1,0,0))
            elif key_sym.lower() == 'z':
                perform_view_reset((0,0,1))
            
            elif key_sym == 'P':
                # Make invisible
                current_vis_text = textActor.GetVisibility()
                current_vis_hint = hintActor.GetVisibility()
                textActor.SetVisibility(False)
                hintActor.SetVisibility(False)

                w2if = vtk.vtkWindowToImageFilter()
                w2if.SetInput(render_window)
                w2if.SetInputBufferTypeToRGB()
                w2if.ReadFrontBufferOff()
                w2if.Update()

                now = datetime.datetime.now()
                timestamp = now.strftime("%Y-%m-%d_%H-%M-%S")

                writer = vtk.vtkPNGWriter()
                writer.SetFileName(f'vtk_{timestamp}.png')
                writer.SetInputConnection(w2if.GetOutputPort())
                writer.Write()
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")

                # Restore visibility
                textActor.SetVisibility(current_vis_text)
                hintActor.SetVisibility(current_vis_hint)

        priority = 1.0
        render_window_interactor.AddObserver(vtk.vtkCommand.CharEvent, key_press_callback, priority)

        # Start
        render_window.Render()
        
        # Camera configuration
        cam = renderer.GetActiveCamera()
        
        # Apply initial projection type
        cam.SetParallelProjection(orthographic)

        # Set default camera position
        if has_system_actors:
            renderer.ResetCamera(sys_bounds)
        else:
            renderer.ResetCamera()

        # Set custom camera position
        if camera_position is not None:
            cam.SetPosition(*camera_position)
            cam.SetViewUp(*view_up)

        # Set custom focal point
        if focal_point is not None:
            if isinstance(focal_point, str):
                if focal_point in self.os:
                    fp = self.os[focal_point].position
                    if self._tel_to_local_rot is not None:
                        fp = self._tel_to_local_rot @ fp + self._translation_vector
                    cam.SetFocalPoint(*fp)
                else:
                    quit(render_window_interactor)
                    raise(RuntimeError(f'Surface {focal_point} not found in optical system.'))
            else:
                cam.SetFocalPoint(*focal_point)
            
            # Recalculate clipping to ensure the new view doesn't cut off objects
            renderer.ResetCameraClippingRange()
        
        # Increase far clip plane
        clip_rng = cam.GetClippingRange()
        cam.SetClippingRange(clip_rng[0], clip_rng[1] * 1000)

        render_window.Render()
        render_window_interactor.Initialize()
        render_window_interactor.Start()

        # Stop
        quit(render_window_interactor)

    def _create_wavelength_lut(self):
        """Creates a color transfer function for wavelength in the range 200nm - 1000nm."""
        ctf = vtk.vtkColorTransferFunction()
        ctf.SetColorSpaceToRGB()
        
        # Bright white/blue
        ctf.AddRGBPoint(200.0, 0.9, 0.9, 1.0) 
        # Electric purple
        ctf.AddRGBPoint(300.0, 0.7, 0.5, 1.0)   
        
        # Standard Rainbow, fully saturated
        ctf.AddRGBPoint(380.0, 0.5, 0.0, 1.0)   # Violet
        ctf.AddRGBPoint(440.0, 0.0, 0.2, 1.0)   # Blue
        ctf.AddRGBPoint(490.0, 0.0, 1.0, 1.0)   # Cyan
        ctf.AddRGBPoint(510.0, 0.2, 1.0, 0.2)   # Green
        ctf.AddRGBPoint(580.0, 1.0, 1.0, 0.0)   # Yellow
        ctf.AddRGBPoint(600.0, 1.0, 0.5, 0.0)   # Orange
        ctf.AddRGBPoint(700.0, 1.0, 0.0, 0.0)   # Red
        
        # Bright cherry red
        ctf.AddRGBPoint(850.0, 0.9, 0.2, 0.2)   
        # Pale pink
        ctf.AddRGBPoint(1000.0, 0.7, 0.5, 0.5)
        
        return ctf

    def add_rays(self, start, stop, surface_id=None, wavelengths=None, directions=None, length=None, point_size=1.0, show_rays=True, show_hits=True):
        """
        Draw rays from start to stop positions and highlight stop points.
        
        If a stop position is NaN (indicating a miss), the ray is skipped by default.
        If 'length' and 'directions' are provided, rays with NaN stops are drawn 
        starting from 'start' along 'direction' for 'length' units (without a hit dot).
        
        Parameters
        ----------
        start : ndarray
            (n, 3) array of starting coordinates (x, y, z).
        stop : ndarray
            (n, 3) array of stopping coordinates (x, y, z).
        surface_id: ndarray
            (n,) array of surface indices reached by each ray.
        directions : ndarray, optional
            (n, 3) array of direction vectors. Required if plotting NaN rays with fixed length.
        length : float, optional
            If provided, rays with NaN stop will be drawn with this length using the direction vector.
        opacity : float, optional
            Transparency of the rays from 0.0 (invisible) to 1.0 (opaque). Default is 0.5.
        point_size : float, optional
            Size of the hit points. Default is 1.0
        show_rays : bool, optional
            Whether to show rays or not. Default is True.
        show_hits : bool, optional
            Whether to show hits or not. Default is True.
        """

        if not show_hits and not show_rays:
            return

        has_directions = directions is not None

        has_wavelengths = wavelengths is not None

        if start.shape != stop.shape:
            raise ValueError("Start and stop points must have the same shape.")

        if has_directions and start.shape != directions.shape:
            raise ValueError("Start/stop poinnts and directions must have the same shape.")
        
        if has_wavelengths and start.shape[0] != wavelengths.shape[0]:
            print(start.shape[0], wavelengths.shape[0])
            raise ValueError("Start/stop points and wavelengths shape mismatch.")
        
        wavelength_scalars = vtk.vtkFloatArray()
        wavelength_scalars.SetName("Wavelengths")
        
        n_photons = start.shape[0]
        
        # Rays
        points_rays = vtk.vtkPoints()
        lines_rays = vtk.vtkCellArray()
        
        # Hits
        hits_collections = {}
        if surface_id is None:
            surface_id = np.zeros((n_photons,), dtype=np.int32)

        # Check which rays have intersected a surface
        valid_stops = ~np.any(np.isnan(stop), axis=1)
        
        for i in range(n_photons):
            current_start = start[i]
            current_stop = stop[i]
            label = self.os[surface_id[i]].name
            
            is_valid_stop = valid_stops[i]
            
            final_stop = None
            draw_hit = False
            if is_valid_stop:
                # Draw from start to stop and add a hit
                final_stop = current_stop
                draw_hit = True
            elif length is not None and has_directions:
                # Draw fixed length ray if directions are available
                final_stop = current_start + directions[i] * length
                draw_hit = False
            else:
                # Ignore not intersected ray
                continue
            
            if show_rays:
                ## Ray logic
                id_start = points_rays.InsertNextPoint(current_start)
                id_stop_line = points_rays.InsertNextPoint(final_stop)
                
                line = vtk.vtkLine()
                line.GetPointIds().SetId(0, id_start)
                line.GetPointIds().SetId(1, id_stop_line)
                lines_rays.InsertNextCell(line)

                # Store wavelength
                if has_wavelengths:
                    wavelength_scalars.InsertNextValue(wavelengths[i])
            
            ## Hit logic
            if draw_hit:
                if label not in hits_collections:
                    hits_collections[label] = {
                        'points': vtk.vtkPoints(),
                        'verts': vtk.vtkCellArray()
                    }
                # Add point to specific surface collection
                pt_id = hits_collections[label]['points'].InsertNextPoint(final_stop)
                hits_collections[label]['verts'].InsertNextCell(1)
                hits_collections[label]['verts'].InsertCellPoint(pt_id)
            
        ## Rays actor
        if show_rays:
            raysPolyData = vtk.vtkPolyData()
            raysPolyData.SetPoints(points_rays)
            raysPolyData.SetLines(lines_rays)
            
            mapper_rays = vtk.vtkPolyDataMapper()
            mapper_rays.SetInputData(raysPolyData)

            if has_wavelengths:
                raysPolyData.GetCellData().SetScalars(wavelength_scalars)                
                # Use the lookup table
                lut = self._create_wavelength_lut()
                mapper_rays.SetLookupTable(lut)
                mapper_rays.SetScalarRange(200, 1000) # nm
                mapper_rays.SetScalarModeToUseCellData()
            else:
                mapper_rays.ScalarVisibilityOff()
            
            actor_rays = vtk.vtkActor()
            actor_rays.SetMapper(mapper_rays)

            if not has_wavelengths:
                actor_rays.GetProperty().SetColor(0.0, 0.7, 1.0) # Electric blue
            
            actor_rays.GetProperty().SetLineWidth(1.0)
            if self._ray_opacity is not None:
                actor_rays.GetProperty().SetOpacity(float(self._ray_opacity))
            
            # Make rays unpickable
            actor_rays.PickableOff()

            base_key = 'rays'
            key = base_key
            count = 0
            while key in self.actors:
                key = f"{base_key}_{count}"
                count += 1
            
            self.actors[key] = actor_rays

        ## Hits actor
        if show_hits:
            for label, data in hits_collections.items():
                if data['points'].GetNumberOfPoints() == 0:
                    continue
            
                dotsPolyData = vtk.vtkPolyData()
                dotsPolyData.SetPoints(data['points'])
                dotsPolyData.SetVerts(data['verts'])
                
                mapper_dots = vtk.vtkPolyDataMapper()
                mapper_dots.SetInputData(dotsPolyData)
                
                actor_dots = vtk.vtkActor()
                actor_dots.SetMapper(mapper_dots)
                actor_dots.GetProperty().SetColor(1.0, 1.0, 0.0)
                actor_dots.GetProperty().SetPointSize(point_size)
                actor_dots.GetProperty().SetOpacity(1.0)
                actor_dots.PickableOff()

                # Naming convention: hits_SurfaceID
                # If multiple batches are added, append an index
                base_key = f"hits_{label}"
                key = base_key
                count = 0
                while key in self.actors:
                    key = f"{base_key}_{count}"
                    count += 1
                
                self.actors[key] = actor_dots

    def _set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        if value is None:
            return
        
        for actor_name in self.actors:
            if actor_name.startswith("rays"):
                self.actors[actor_name].GetProperty().SetOpacity(float(value))

    def set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        self._ray_opacity = value
        self._set_ray_opacity(value)

    def _apply_global_transform(self):
        """
        Apply a global rotation and translation to all current actors.
        
        Parameters
        ----------
        R : ndarray
            (3, 3) Rotation matrix. 
        t : ndarray or list
            (3,) Translation vector (x, y, z).
        """
        # Convert numpy array to vtkMatrix4x4
        vtk_R = vtk.vtkMatrix4x4()
        for i in range(3):
            for j in range(3):
                vtk_R.SetElement(i, j, self._tel_to_local_rot[i, j])

        # Iterate over all actors
        for actor in self.actors.values():
            
            # Get the existing transform
            transform = actor.GetUserTransform()
            if transform is None:
                transform = vtk.vtkTransform()
                transform.SetMatrix(actor.GetMatrix())
                actor.SetUserTransform(transform)
            
            # Apply transformations
            # PostMultiply ensures we are applying this to the "Global" world coordinates
            transform.PostMultiply()
            
            # Apply rotation
            transform.Concatenate(vtk_R)
            
            # Apply translation
            transform.Translate(*self._translation_vector)

            transform.Update()

    def get_screenshot(self,
        camera_position,
        filename=None,
        focal_point=None,
        view_up=(0, 0, 1),
        size=None,
        image_scale=1,
        orthographic=False,
        representation='surface',
        hide_surfaces=None,
        zoom=1.0,
        show=False,
        position_reference='telescope',
        position_shift=(0,0,0)
    ) -> np.ndarray:
        """
        Render the scene off-screen and save it to a PNG file.

        Coordinates for camera position, focal point, and view up should be provided 
        in the telescope coordinate system. The method handles the transformation 
        to the local coordinate system automatically.

        Parameters
        ----------
        camera_position : sequence of float
            The (x, y, z) coordinates of the camera position. The reference frame can be specified with the position_reference parameter.
        filename : str, optional
            Path to save the resulting PNG file. If None, the screenshot is not saved.
        focal_point : str or sequence of float, optional
            The point the camera is looking at. 
            - If None (default): The camera looks at the center of the bounding box of all system actors.
            - If str: Looks for an optical surface with this name and targets its position.
            - If sequence: Uses the provided (x, y, z) coordinates (assumend to be local coordinates).
        view_up : sequence of float
            The vector defining the "up" direction for the camera, default=(0, 0, 1).
        size : tuple of int, optional
            The (width, height) of the rendered image in pixels. 
            If None, uses the current window size.
        image_scale : int
            Resolution multiplier (e.g., 2 doubles the pixel density), default=1.
        orthographic : bool
            If True, uses parallel projection.
            If False, uses standard perspective projection.
            By default False.
        representation : {'surface', 'wireframe', 'points'}
            The rendering style of the actors in the scene, default='surface'.
        hide_surfaces : list of str, optional
            A list of names of surfaces/actors to exclude from the render.
        zoom : float
            Zoom factor applied when ``orthographic`` is True.
            - 1.0: Fits the object diagonal to the screen height.
            - > 1.0: Zooms in.
            - < 1.0: Zooms out.
            By default 1.0.
            Note: This parameter currently implies no effect if orthographic is False.
        show : bool
            If True, displays the image using Matplotlib. Default is False.
        position_reference:
            Reference system of the specified camera position, by default 'telescope'.
            If does not starts with 't', the camera position is assumed to be relative to the local reference system.
        
        Returns
        -------
        numpy.ndarray
            RGB image array with shape (height, width, 3).
        """
        # Setup Renderer
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)
        render_window = vtk.vtkRenderWindow()
        render_window.SetOffScreenRendering(1)
        render_window.AddRenderer(renderer)
        
        if size: render_window.SetSize(*size)
        else: render_window.SetSize(*self.window_size)

        # Ray Opacity
        if self._ray_opacity is None:
            n_rays = 0
            for actor_name, actor in self.actors.items():
                if actor_name.startswith('rays'):
                    n_rays += actor.GetMapper().GetDataSetInput().GetLines().GetNumberOfCells()
            self._set_ray_opacity(self._get_sigmoid_opacity(n_rays))

        # Add actors and calculate bounds
        inf = float('inf')
        sys_bounds = [inf, -inf, inf, -inf, inf, -inf]
        has_system_actors = False
        hidden_set = set(hide_surfaces) if hide_surfaces else set()

        for name, actor in self.actors.items():
            # Check if the surface must be hidden
            should_skip = name in hidden_set
            if not should_skip and name.startswith("hits_"):
                for hidden_name in hidden_set:
                    if name.startswith(f"hits_{hidden_name}"):
                        should_skip = True
                        break
            if should_skip:
                continue

            # Set representation
            prop = actor.GetProperty()
            if representation == 'wireframe':
                prop.SetRepresentationToWireframe()
            elif representation == 'points':
                prop.SetRepresentationToPoints()
            else: prop.SetRepresentationToSurface()

            renderer.AddActor(actor)
            
            # Update bounds (using world coordinates of the actor)
            if not name.startswith("rays"):
                has_system_actors = True
                b = actor.GetBounds()
                sys_bounds[0] = min(sys_bounds[0], b[0])
                sys_bounds[1] = max(sys_bounds[1], b[1])
                sys_bounds[2] = min(sys_bounds[2], b[2])
                sys_bounds[3] = max(sys_bounds[3], b[3])
                sys_bounds[4] = min(sys_bounds[4], b[4])
                sys_bounds[5] = max(sys_bounds[5], b[5])

        # Configure camera coordinates
        cam = renderer.GetActiveCamera()
        cam.SetParallelProjection(orthographic)

        # Handle focal point
        final_fp = np.zeros(3)
        if focal_point is not None:
            if isinstance(focal_point, str):
                # Look for surface by name
                if focal_point not in self.os:
                    raise(RuntimeError(f"Surface {focal_point} not found."))
                s = self.os[focal_point]
                fp_local = np.array(s.position, dtype=float)
                if self._tel_to_local_rot is not None:
                    final_fp = self._tel_to_local_rot @ fp_local + self._translation_vector
                else:
                    final_fp = fp_local
            else:
                fp_input = np.array(focal_point, dtype=float)
                final_fp = fp_input
        elif has_system_actors:
            # Center of bounding box
            final_fp[0] = (sys_bounds[0] + sys_bounds[1]) / 2.0
            final_fp[1] = (sys_bounds[2] + sys_bounds[3]) / 2.0
            final_fp[2] = (sys_bounds[4] + sys_bounds[5]) / 2.0

        # Camera position
        final_pos = np.array(camera_position, dtype=float)
        if self._tel_to_local_rot is not None and position_reference.startswith('t'):
            final_pos = self._tel_to_local_rot @ final_pos + self._translation_vector

        # Camera shift
        view_vec = final_fp - final_pos
        shift_vec = np.array(position_shift, dtype=float)
        if self._tel_to_local_rot is not None and position_reference.startswith('t'):
            shift_vec = self._tel_to_local_rot @ shift_vec
        final_pos += shift_vec
        final_fp = final_pos + view_vec

        # Set camera parameters
        cam.SetPosition(*final_pos)
        cam.SetViewUp(*view_up)
        cam.SetFocalPoint(*final_fp)

        # Handle orthographic scaling
        if orthographic and has_system_actors:
            # Calculate diagonal length of the bounding box
            dx = sys_bounds[1] - sys_bounds[0]
            dy = sys_bounds[3] - sys_bounds[2]
            dz = sys_bounds[5] - sys_bounds[4]
            diagonal = np.sqrt(dx*dx + dy*dy + dz*dz)
            
            # Set scale so the whole object fits, adjusted by zoom
            # ParallelScale is half the height of the viewport
            if zoom == 0:
                zoom = 1.0
            cam.SetParallelScale((diagonal / 2.0) / zoom)

        renderer.ResetCameraClippingRange()
        render_window.Render()

        # Capture
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput(render_window)
        w2if.SetScale(image_scale)
        w2if.SetInputBufferTypeToRGB()
        w2if.ReadFrontBufferOff()
        w2if.Update()
        
        # Save
        if filename:
            writer = vtk.vtkPNGWriter()
            writer.SetFileName(filename)
            writer.SetInputConnection(w2if.GetOutputPort())
            writer.Write()

        from vtkmodules.util.numpy_support import vtk_to_numpy
        img_data = w2if.GetOutput()
        width, height, _ = img_data.GetDimensions()
        sc = img_data.GetPointData().GetScalars()
        arr = vtk_to_numpy(sc)
        arr = arr.reshape(height, width, -1)
        arr = np.flipud(arr)
        
        # Show
        if show:
            import matplotlib.pyplot as plt
            plt.imshow(arr)
            plt.axis('off')
            plt.show()

        return arr


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
import numpy as np

from ..visualization._iactsim_style import iactsim_style

@dataclass
class AtmosphericTransmission:
    """
    Represents atmospheric optical depth as a function of photon wavelength and emission altitude.

    Attributes:
        value: A numpy.ndarray of shape (n,m) representing the atmosphere transmission (from 0 to 1).
        wavelength: A numpy.ndarray of shape (n,) representing the wavelengths in nanometers.
        altitude: A numpy.ndarray of shape (m,) representing the emission altitude in millimeters.
    """
    value: np.ndarray = field(default=None, metadata={'units': 'None'})
    wavelength: np.ndarray = field(default=None, metadata={'units': 'nanometers'})
    altitude: np.ndarray = field(default=None, metadata={'units': 'millimeters'})

    def __repr__(self):
        field_info = []
        for field_name, field_value in self.__dict__.items():
            if field_value is None:
                field_info.append(f"{field_name}: not set")
            else:
                field_info.append(f"{field_name}: set")
        return f"AtmosphericTransmission({', '.join(field_info)})"
    
    @iactsim_style
    def plot(self, heatmap=None, depth=True):
        if self.wavelength is None or self.altitude is None or self.value is None:
            raise(RuntimeError("Atmosphere transmission has not been initialized properly."))
        
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()

        n_wl = len(self.wavelength)
        n_th = len(self.altitude)
        
        if heatmap is None:
            heatmap = True if n_th > 10 else False
        
        data = self.value.copy()

        ylabel = 'Optical Depth'
        if not depth:
            data = np.exp(-data)
            ylabel = 'Transmission'

        if not heatmap:
            data = data.flatten()
            for i in range(n_th):
                label = None
                label = f'{self.altitude[i]/1e6:.1f} km'
                plt.plot(self.wavelength, data[i*n_wl:(i+1)*n_wl], label=label)
            if n_th < 10:
                plt.legend()
            plt.xlabel('Wavelength (nm)')
            plt.ylabel(ylabel)
            if depth:
                plt.yscale('log')
            ax.grid(which='both')
        else:
            levels = np.linspace(0,1,100) if not depth else np.logspace(np.log10(data[data>0].min()),np.log10(data[data<99999].max()),100)
            contour = ax.contourf(self.wavelength, self.altitude/1e6, data, levels=levels)
            ax.set_xlabel("Wavelength (nm)")
            ax.set_ylabel("Emission altitude (km)")
            fig.colorbar(contour, ax=ax, boundaries=[0,1])

        return fig, ax



# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
import numpy as np
from tabulate import tabulate
from typing import Tuple, Optional, Any
from numpy.typing import ArrayLike

from ._sag import sag
from ._surface_misc import SurfaceType, ApertureShape, SurfaceProperties, SurfaceShape, SensorSurfaceType
from ._materials import Material, Materials

@dataclass
class SurfaceVisualProperties:
    """
    Stores visualization attributes for an optical surface.

    Parameters
    ----------
    color : tuple of float, optional
        RGB color tuple (0.0-1.0). If None, the visualizer uses default logic based on surface type.
    opacity : float, optional
        Opacity (0.0 transparent - 1.0 opaque). If None, visualizer uses default logic based on surface type.
    specular : float, optional
        Specular reflection intensity (0.0-1.0).
    wireframe : bool, optional
        If True, forces wireframe representation for this surface. Default is False.
    resolution : float, optional
        Mesh resolution in mm. If None, uses the visualizer logic based on the surface extent.
    visible : bool, optional
        Whether the surface is visible in the render. Default is True.
    """
    color: Optional[Tuple[float, float, float]] = None
    opacity: Optional[float] = None
    specular: Optional[float] = None
    wireframe: bool = False
    resolution: Optional[float] = None
    visible: bool = True

class Surface:
    """Represents an optical surface.

        Parameters
        ----------
        type : SurfaceType
            Type of surface (reflective, refractive, opaque, dummy)
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        visual_properties : SurfaceVisualProperties, optional
            Visual properties of the surface, by default None
        name : str, optional
            Surface name
    """

    def __init__(self,
                 surface_type: SurfaceType,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 visual_properties: Optional['SurfaceVisualProperties'] = None,
                 name: Optional[str] = None):
        self.type = surface_type
        self.position = position
        self.tilt_angles = tilt_angles
        self.scattering_dispersion = scattering_dispersion
        self.properties = properties if properties is not None else SurfaceProperties()
        self.material_in = material_in
        self.material_out = material_out
        self.name = name
        self.sensor_type = SensorSurfaceType.NONE
        if visual_properties is not None:
            self.visual_properties = visual_properties
        else:
            self.visual_properties = SurfaceVisualProperties()

        self._custom_rotation_matrix = None

    @property
    def axial_position(self) -> float:
        return self._axial_position

    @axial_position.setter
    def axial_position(self, value: float):
        self._axial_position = value
        self._position[-1] = value

    @property
    def position(self) -> np.ndarray:
        return self._position

    @position.setter
    def position(self, value: np.ndarray):
        self._position = np.asarray(value)
        self._axial_position = value[-1]
    
    def get_rotation_matrix(self) -> np.ndarray:
        """
        Calculates the 3D rotation matrix to transform into the surface reference
        system from the telescope reference system.
        
        The computation is based on surface tilt angles and assumes rotations are applied in the order 
        (last first) Z @ Y @ X. This corresponds to rotating the surface around 
        the telescope X axis, then the telescope Y axis and finally the telescope Z axis.
        The rotations follow the right-hand rule (counter-clockwise).

        These angles are intended for small misalignments,
        while :py:func:`~iactsim.optics.transforms.create_surface_rotation`
        is a more robust method. Tilt angles can be bypassed if 
        a custom rotation matrix has been set using :py:meth:`set_rotation_matrix` method.
        
        Tilt angles are expected in degrees. 

        Returns
        -------
            np.ndarray: A 3x3 rotation matrix.
            The returned matrix is the transpose of the tilt sequence. It transforms from
            the telescope reference system into the surface reference system.
        
        """
        # Return the custom rotation matrix if has been defined
        if self._custom_rotation_matrix is not None:
            return self._custom_rotation_matrix.reshape(3,3)

        x_tilt_rad, y_tilt_rad, z_tilt_rad = np.deg2rad(self.tilt_angles)
        cos_x = np.cos(x_tilt_rad)
        sin_x = np.sin(x_tilt_rad)
        cos_y = np.cos(y_tilt_rad)
        sin_y = np.sin(y_tilt_rad)
        cos_z = np.cos(z_tilt_rad)
        sin_z = np.sin(z_tilt_rad)

        x_rot = np.array([
            [1.0,    0.,     0.],
            [ 0., cos_x, -sin_x],
            [ 0., sin_x,  cos_x]
        ])

        y_rot = np.array([
            [ cos_y,  0., sin_y],
            [    0.,  1.,    0.],
            [-sin_y,  0., cos_y]
        ])

        z_rot = np.array([
            [ cos_z, -sin_z, 0.],
            [ sin_z,  cos_z, 0.],
            [    0.,     0., 1.],
        ])

        return  (z_rot @ y_rot @ x_rot).T
    
    def set_rotation_matrix(self, a_matrix):
        """Set a custom rotation matrix instead of deriving it from tilt angles.
        To use tilt angles set it to ``None``.

        Parameters
        ----------
        a_matrix : np.ndarray
            3D rotation matrix to transform from telescope to surface reference frame.
        """
        self._custom_rotation_matrix = a_matrix
    
    def __repr__(self) -> str:
        """Returns a string representation of the surface object."""
        return f"{self.__class__.__name__}(half_aperture={self.type}, ...)"
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
            ["Scattering Sigma", self.scattering_dispersion],
            ["Properties wavelegth", "set" if self.properties.wavelength is not None else "not set"],
            ["Properties incidence angle", "set" if self.properties.incidence_angle is not None else "not set"],
            ["Transmittance value", "set" if self.properties.transmittance is not None else "not set"],
            ["Reflectance value", "set" if self.properties.reflectance is not None else "not set"],
            ["Absorption value", "set" if self.properties.absorption is not None else "not set"],
        ]
        return table

    def __str__(self) -> str:
        """Returns a string representation of the surface object formatted as a table."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="fancy_grid")
    
    def _repr_markdown_(self) -> str:
        """Returns a Markdown representation of the surface object."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="pipe")

    def _repr_html_(self) -> str:
        """Returns an HTML representation of the surface object."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="html")

class AsphericalSurface(Surface):
    """Represents an aspherical optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        curvature : float
            Surface curvature (1/radius of curvature)
        conic_constant : float
            Conic constant
        aspheric_coefficients : array-like
            Array of aspheric coefficients
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        is_fresnel : bool, optional
            Whether the surface is an ideal Fresnel lens/mirror, by default False
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 curvature: float,
                 conic_constant: float,
                 aspheric_coefficients: ArrayLike,
                 central_hole_half_aperture: float = 0.0,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 is_fresnel: bool = False,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self.half_aperture = half_aperture
        self.curvature = curvature
        self.conic_constant = conic_constant
        self.aspheric_coefficients = np.asarray(aspheric_coefficients)
        self.central_hole_half_aperture = central_hole_half_aperture
        self.central_hole_shape = central_hole_shape
        self.aperture_shape = aperture_shape
        self.is_fresnel = is_fresnel
        self.offset = np.array([0.,0.])
        """Treat the surface as a segment centered at these aperture coordinates."""
        self._shape = SurfaceShape.ASPHERICAL

    def _validate(self):
        """Validates the internal consistency of the surface parameters."""
        name_str = "-".join([self.name,""]) if self.name is not None else ""
        if not self.half_aperture > 0:
            raise ValueError("".join([name_str, "Half-aperture must be positive."]))
        if not self.central_hole_half_aperture >= 0:
            raise ValueError("".join([name_str, "Half-aperture of the central hole cannot be negative."]))
        if not self.central_hole_half_aperture < self.half_aperture:
            raise ValueError("".join([name_str, "Half-aperture of the central must be less than surface half-aperture."]))
        if not self.aspheric_coefficients.ndim == 1:
            raise ValueError("".join([name_str, "Aspheric coefficients must be a 1D array."]))

    @property
    def aspheric_coefficients(self) -> np.ndarray:
        return self._aspheric_coefficients

    @aspheric_coefficients.setter
    def aspheric_coefficients(self, value: np.ndarray):
        if not isinstance(value, np.ndarray):
            value = np.asarray(value, dtype=np.float64)
        if len(value) > 10:
            raise(ValueError("Aspheric surfaces can be defined by a maximum of 10 aspheric coefficients."))
        self._aspheric_coefficients = value

    def sagitta(self, r: np.ndarray) -> np.ndarray:
        """Calculate the sagitta of the aspheric surface at a given radial distance in the surface refrence frame.

        Parameters
        ----------
        r : np.ndarray, list or scalar
            Radial distance(s) from the optical axis.

        Returns
        -------
        np.ndarray
            Sagitta value(s) at the given radial distance(s).
        """

        return sag(r, self.curvature, self.conic_constant, self.aspheric_coefficients, 0., self.is_fresnel)

    def get_surface_normal_vector(self, x: float, y: float) -> np.ndarray:
        """Calculate the upward-going normal unit vector at a given surface point (x,y) in the surface reference frame.

        Parameters
        ----------
        x : float
            x coordinate of the point.
        y : float
            y coordinate of the point.

        Returns
        -------
        np.ndarray
            Normal unit vector.
        """
        n = np.array([0,0,1.])

        # Flat
        if (np.abs(self.curvature) < 1e-10): 
            return n
        
        r2 = x*x+y*y

        # Aspheric component
        tot_aspher = 0.
        for i in range(len(self.aspheric_coefficients)):
            tot_aspher += 2 * (i+1) * self.aspheric_coefficients[i] * np.power(r2, i)

        arg_sqrt = 1-(1.0 + self.conic_constant) * self.curvature*self.curvature * r2
        if arg_sqrt < 0:
            raise(ValueError('Surface not well defined at the given point.'))
        
        fact = self.curvature / np.sqrt(arg_sqrt) + tot_aspher

        n[0] = - fact * x
        n[1] = - fact * y
        norm = np.sqrt(n[0]*n[0]+n[1]*n[1]+1.)

        return n/norm
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Half-Aperture", f'{self.half_aperture:.2f}'],
            ["Curvature", f'{self.curvature:.9f}'],
            ["Conic Constant", f'{self.conic_constant:.2f}'],
            ["# Aspheric Coefficients", len(self.aspheric_coefficients)],
            ["Central Hole Half-Aperture", f'{self.central_hole_half_aperture:.2f}'],
            ["Aperture Shape", self.aperture_shape],
            ["Central Hole Shape", self.central_hole_shape],
            ["Is Fresnel", self.is_fresnel],
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
            ["Scattering Sigma", f'{self.scattering_dispersion:.3f}'],
            ["Properties wavelegth", "set" if self.properties.wavelength is not None else "not set"],
            ["Properties incidence angle", "set" if self.properties.incidence_angle is not None else "not set"],
            ["Transmittance value", "set" if self.properties.transmittance is not None else "not set"],
            ["Reflectance value", "set" if self.properties.reflectance is not None else "not set"],
            ["Absorption value", "set" if self.properties.absorption is not None else "not set"],
        ]
        return table

class FlatSurface(AsphericalSurface):
    """Represents a flat optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 central_hole_half_aperture: float = 0.0,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            half_aperture=half_aperture,
            curvature=0,
            conic_constant=0,
            aspheric_coefficients=[0.],
            central_hole_half_aperture=central_hole_half_aperture,
            aperture_shape=aperture_shape,
            central_hole_shape=central_hole_shape,
            is_fresnel=False,
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self._shape = SurfaceShape.FLAT

class SphericalSurface(AsphericalSurface):
    """Represents a spherical optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        curvature : float
            Surface curvature (1/radius of curvature)
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        is_fresnel : bool, optional
            Whether the surface is an ideal Fresnel lens/mirror, by default False
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 curvature: float,
                 central_hole_half_aperture: float = 0.,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 is_fresnel: bool = False,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            half_aperture=half_aperture,
            curvature=curvature,
            conic_constant=0,
            aspheric_coefficients=[0.],
            central_hole_half_aperture=central_hole_half_aperture,
            aperture_shape=aperture_shape,
            central_hole_shape=central_hole_shape,
            is_fresnel=is_fresnel,
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self._shape = SurfaceShape.SPHERICAL


class CylindricalSurface(Surface):
    """Represents a cylindrical optical surface.

        Parameters
        ----------
        radius : float
            Radius of the cylinder.
        height : float
            Height of the cylinder.
        has_top : bool, optional
            Whether the cylinder has a top cap, by default True.
        has_bottom : bool, optional
            Whether the cylinder has a bottom cap, by default True.
        position : array-like, optional
            Position of cylinder center [x, y, z] (half-height on the cylinder axis), by default None.
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        surface_type : SurfaceType, optional
            Type of surface (opaque or sensitive), by default SurfaceType.OPAQUE.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 radius: float,
                 height: float,
                 has_top: bool = True,
                 has_bottom: bool = True,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 surface_type: SurfaceType = SurfaceType.OPAQUE,
                 name: Optional[str] = None):

        scattering_dispersion = 0.
        properties = None
        material_in = Materials.AIR
        material_out = Materials.AIR
        super().__init__(
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self.radius = radius
        self.height = height
        self.top = has_top
        self.bottom = has_bottom
        self._shape = SurfaceShape.CYLINDRICAL
    
    def _validate(self):
        """Validates the internal consistency of the surface parameters."""
        name_str = "-".join([self.name,""]) if self.name is not None else ""
        if not self.radius > 0:
            raise ValueError("".join([name_str, "Cylinder radius must be positive."]))
        if not self.height > 0:
            raise ValueError("".join([name_str, "Cylinder height must be positive."]))
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Radius", f'{self.radius:.2f}'],
            ["Height", f'{self.height:.2f}'],
            ["Has top", self.top],
            ["Has bottom", self.bottom],
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
        ]
        return table


class SipmTileSurface(FlatSurface):
    """Represents a SiPM tile of NxN pixels.

        Parameters
        ----------
        pixels_per_side : int
            Number of pixels per side.
        pixel_active_side : float
            Side length of the active area of a single pixel.
        pixels_separation : float
            Separation between pixels.
        border_to_active_area : float
            Width of the border from the edge of the tile to the active area.
        microcell_size : float
            Size of a pixel micro-cell in mm (e.g. 0.075 for 75um ucells).
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (counter-clockwise x-tilt, y-tilt and z_tilt) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
        
        Notes
        -----
        Tile layout::

              y       pixels_separation
              ^    ________ <--> ________
              |   |        |    |        |
              |   |        |    |        |
              |   |        |    |        |
              |   |________|    |________|
              |<->               <------>
              | |            pixel_active_side
              | border_to_active_area
            (0,0)---------> x

    """

    def __init__(self,
                 pixels_per_side: int,
                 pixel_active_side: float,
                 pixels_separation: float,
                 border_to_active_area: float,
                 microcell_size: float = 0.0,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 surface_type: SurfaceType = SurfaceType.SENSITIVE,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None
                ):
        
        # Compute tile size
        tile_side = pixels_per_side * pixel_active_side + (pixels_per_side - 1) * pixels_separation
        half_aperture = 0.5 * tile_side + border_to_active_area

        super().__init__(
            half_aperture = half_aperture,
            central_hole_half_aperture = 0.0,
            aperture_shape = ApertureShape.SQUARE,
            surface_type = SurfaceType.SENSITIVE_BACK,
            position = position,
            tilt_angles = tilt_angles,
            scattering_dispersion = scattering_dispersion,
            properties = properties,
            material_in = material_in,
            material_out = material_out,
            name = name
        )
        self.type = surface_type
        self._pixel_active_side = pixel_active_side
        self._pixels_separation = pixels_separation
        self._border_to_active_area = border_to_active_area
        self._pixels_per_side = pixels_per_side
        self._n_pixels = pixels_per_side * pixels_per_side

        # By deafult no ucells
        self._n_ucells_per_side = 0
        self.microcell_size = microcell_size

        self.sensor_type = SensorSurfaceType.SIPM_TILE
        self._shape = SurfaceShape.FLAT
        self.sensor_id = 0
        self._first_pixel_id = None
    
    @property
    def type(self) -> SurfaceType:
        return self._type

    @type.setter
    def type(self, value: SurfaceType):
        sensitive_list = [
            SurfaceType.SENSITIVE_BACK,
            SurfaceType.SENSITIVE_FRONT,
            SurfaceType.SENSITIVE,
            SurfaceType.REFLECTIVE_SENSITIVE,
            SurfaceType.REFLECTIVE_SENSITIVE_FRONT,
            SurfaceType.REFLECTIVE_SENSITIVE_BACK,
        ]
        if value not in sensitive_list:
            raise(ValueError("SipmTileSurface can be only a sensitive surface."))
        self._type = value
    
    @property
    def pixels_per_side(self) -> int:
        return self._pixels_per_side

    @pixels_per_side.setter
    def pixels_per_side(self, value: int):
        self._pixels_per_side = value
        self._n_pixels = value * value
        self._update_size()
    
    @property
    def pixel_active_side(self) -> float:
        return self._pixel_active_side
    
    @pixel_active_side.setter
    def pixel_active_side(self, value: float):
        self._pixel_active_side = value
        self._update_size()
    
    @property
    def microcell_size(self):
        return self._microcell_size
    
    @microcell_size.setter
    def microcell_size(self, value: float):
        self._microcell_size = value
        if value > 0:
            ucelss_per_side = np.floor(self.pixel_active_side / value)
            self._n_ucells_per_side = int(ucelss_per_side)
            self._pixel_active_side = ucelss_per_side * value
            self._update_size()
            self._inv_ucell_size = 1. / value
        else:
            self._n_ucells_per_side = 0
            self._inv_ucell_size = 0.

    @property
    def pixels_separation(self) -> float:
        return self._pixels_separation
    
    @pixels_separation.setter
    def pixels_separation(self, value: float):
        self._pixels_separation = value
        self._update_size()
    
    @property
    def border_to_active_area(self) -> float:
        return self._border_to_active_area
    
    @border_to_active_area.setter
    def border_to_active_area(self, value: float):
        self._border_to_active_area = value
        self._update_size()
    
    def _update_size(self):
        tile_side = self.pixels_per_side * self.pixel_active_side + (self.pixels_per_side - 1) * self.pixels_separation
        self.half_aperture = 0.5 * tile_side + self.border_to_active_area


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import abc
import numpy as np
from typing import Union, List

class ModularSiPMCameraGeometry(abc.ABC):
    """
    Abstract base class for defining the geometry of the focal plane of a Silicon Photomultiplier (SiPM) camera of squared SiPM tiles.
    Default unit is mm.
    """

    @property
    @abc.abstractmethod
    def body_dimensions(self) -> Union[tuple[float], List[float], np.ndarray]:
        """Camera body dimensions:
        
            - width, height (optical axis), depth for rectangular cameras;
            - radius, height (optical axis) for cylindrical cameras.

        Returns
        -------
        Union[tuple[float, float, float], List[float], numpy.ndarray]
            A tuple, list or NumPy array containing the width, height, and depth of the camera body.
            Using numpy array is preferable.
        """
        pass

    @property
    @abc.abstractmethod
    def position(self) -> Union[tuple[float, float, float], List[float], np.ndarray]:
        """Camera surface center position in the focal surface reference frame.

        Returns
        -------
        Union[tuple[float, float, float], List[float], numpy.ndarray]
            A tuple, list or NumPy array representing the (x, y, z) coordinates of the camera surface center.
            Using numpy array is preferable.
        """
        pass

    @property
    @abc.abstractmethod
    def modules_p(self) -> Union[tuple[float, float, float], List[float], np.ndarray]:
        """Modules center position in the focal surface reference frame.

        Returns
        -------
        Union[tuple[float, float, float], List[float], numpy.ndarray]
            A tuple, list or NumPy array representing the (x, y, z) coordinates of the modules center.
            Using numpy array is preferable.
        """
        pass

    @property
    @abc.abstractmethod
    def modules_n(self) -> Union[tuple[float, float, float], List[float], np.ndarray]:
        """Modules normal vector in the focal surface reference frame.

        Returns
        -------
        Union[tuple[float, float, float], List[float], numpy.ndarray]
            A tuple, list or NumPy array representing the (x, y, z) components of the modules normal vector.
            Using numpy array is preferable.
        """
        pass

    @property
    @abc.abstractmethod
    def module_side(self) -> float:
        """Length of module edges from the first to the last pixel active area::

            y
            ^   __   __   __
            |  |__| |__| |__|
            |   __   __   __
            |  |__| |__| |__|
            |   __   __   __
            |  |__| |__| |__|
            |   <---------->
            |   3x3 module side
            -------------------> x

        Returns
        -------
        float
            The length of the module side.
        """
        pass

    @property
    @abc.abstractmethod
    def pixel_active_side(self) -> float:
        """Length of the pixel active area edge::

            y  Active area edge
            ^    <------>
            |    ________      ________
            |   | active |    |        |
            |   |  area  |    |        |
            |   |        |    |        |
            |   |________|    |________|
            |
            --------->x

        Returns
        -------
        float
            The length of the pixel active area side.
        """
        pass

    @property
    @abc.abstractmethod
    def pixels_separation(self) -> float:
        """Gap between pixels active area::

            y       Pixels separation
            ^    ________ <--> ________
            |   |        |    |        |
            |   |        |    |        |
            |   |        |    |        |
            |   |________|    |________|
            |
            --------->x

        Returns
        -------
        float
            The distance between pixel active areas.
        """
        pass


import numpy as np
import pytest
from numpy.testing import assert_allclose

from iactsim.optics._cpu_transforms import (
    pointing_dir,
    local_to_pointing_rotation,
    local_to_telescope_rotation
)

class TestPointingDirection:
    """Tests for `pointing_dir` function."""

    @pytest.mark.parametrize("alt, az, expected_vec", [
        (0, 0,    [1, 0, 0]),  # North -> Local X
        (0, 90,   [0, -1, 0]), # East -> Local -Y
        (0, 180,  [-1, 0, 0]), # South -> Local -X
        (0, 270,  [0, 1, 0]),  # West -> Local +Y
        (90, 0,   [0, 0, 1]),  # Zenith -> Local Z
    ])

    def test_cardinal_directions(self, alt, az, expected_vec):
        """Verify standard cardinal points match the North-West-Up logic."""
        vec = pointing_dir(alt, az)
        assert_allclose(vec, expected_vec, atol=1e-7)

    def test_unit_norm(self):
        """Ensure the output is always a unit vector."""
        for alt in [0, 30, 45, 90]:
            for az in [0, 45, 90, 180, 270]:
                vec = pointing_dir(alt, az)
                assert_allclose(np.linalg.norm(vec), 1.0, atol=1e-7)


class TestRotationMatrices:
    """Tests for rotation matrix generation functions."""

    def test_default_is_identity(self):
        """
        Verify that if no custom_rotation is provided, the function 
        behaves exactly like local_to_pointing_rotation.
        """
        alt, az = 45, 45
        R_default = local_to_telescope_rotation(alt, az)
        R_pointing = local_to_pointing_rotation(alt, az)
        
        assert_allclose(R_default, R_pointing, atol=1e-15,
                        err_msg="Default behavior should be identity (match pointing rotation)")

    def test_astri_rotation(self):
        """
        Verify the specific ASTRI refrence system:
        Target frame when pointing North: 
          - X = West
          - Y = Up
          - Z = North
        """
        # The ASTRI custom transformation matrix pointing North
        astri_matrix = np.array([
            [ 0, 1, 0], # New X (West)
            [-1, 0, 0], # New Y (Up)
            [ 0, 0, 1]  # New Z (North)
        ])
        
        # Test at Reference Position (North, Horizon)
        alt, az = 0, 0
        R = local_to_telescope_rotation(alt, az, custom_rotation=astri_matrix)

        # Local North (1, 0, 0) should become telescope Z (0, 0, 1)
        assert_allclose(R @ [1, 0, 0], [0, 0, 1], atol=1e-7)

        # Local West (0, 1, 0) should become telescope X (1, 0, 0)
        assert_allclose(R @ [0, 1, 0], [1, 0, 0], atol=1e-7)

        # Local Up (0, 0, 1) should become telescope Y (0, 1, 0)
        assert_allclose(R @ [0, 0, 1], [0, 1, 0], atol=1e-7)

    def test_orthogonality(self):
        """Rotation matrices must be orthogonal."""
        alt, az = 30, 60
        R = local_to_telescope_rotation(alt, az) # Uses default identity
        
        identity = np.eye(3)
        assert_allclose(R @ R.T, identity, atol=1e-7)
        assert_allclose(np.linalg.det(R), 1.0, atol=1e-7)

    def test_pointing_frame_orientation_default(self):
        """
        Verify the orientation of the default output (pointing frame).
        
        At Alt=0, Az=0 (pointing North):
        - Z axis points North (local X)
        - X axis points Down (local -Z)
        - Y axis points West (local Y)
        """
        R = local_to_telescope_rotation(0, 0) # Default
        
        # Local North (1, 0, 0) -> pointing Z (0, 0, 1)
        assert_allclose(R @ [1, 0, 0], [0, 0, 1], atol=1e-7)

        # Local down (0, 0, -1) -> pointing X (1, 0, 0)
        assert_allclose(R @ [0, 0, -1], [1, 0, 0], atol=1e-7)

        # Local West (0, 1, 0) -> pointing Y (0, 1, 0)
        assert_allclose(R @ [0, 1, 0], [0, 1, 0], atol=1e-7)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib import colors
import matplotlib.pyplot as plt
import numpy as np
import cupy as cp

coordinate_indices = {
    'X': 0, 'x': 0,
    'Y': 1, 'y': 1,
    'Z': 2, 'z': 2,
}

def histogram2d(vectors3d, bins, plane='xy', to_weight=None, dx=None, ax=None, scale=1., norm=None, range=None, update_this=None, log=False, vmin=None, vmax=None, cmap='viridis', interpolation='nearest', colorbar=True, cbar_label=None, title=None, xlabel=None, ylabel=None, **kwargs):
    """Generates and displays a 2D histogram of two components of a set of 3D vectors.

    Parameters
    ----------
    vectors3d : numpy.ndarray
        A NumPy array representing a set of 3D vectors. The first two columns are used as x and y coordinates for the histogram. The third column is ignored.
    bins : int or array_like
        The number of bins or the bin edges (see `numpy.histogram2d` documentation for details).
    plane : str
        String representation of the components to plot, e.g.: 'xy', 'zy', 'zx', etc.
    to_weight : numpy.ndarray, optional
        Weights for each vector. If None, all vectors are weighted equally.
    dx : float, optional
        Determines the range if `range` is not explicitly provided. Defines a square region around the data's center with sides of length 2*dx.
    ax : matplotlib.axes.Axes, optional
        An Axes object to plot on. If None, the current axes (plt.gca()) is used.
    scale : float, optional
        Scaling factor applied to x and y coordinates.
    norm : matplotlib.colors.Normalize, optional
        Normalization instance for histogram data (e.g., `colors.LogNorm` for logarithmic scaling).
    range : sequence, optional
        A sequence of (min, max) pairs for each dimension, specifying the data range to be histogrammed.
    update_this : numpy.ndarray, optional
        A 2D array (another histogram) to be added to the calculated histogram.
    log : bool, optional
        If True, applies a log1p transformation (log(1 + x)) to the histogram counts.
    vmin : float, optional
        Minimum value for colormap scaling.
    vmax : float, optional
        Maximum value for colormap scaling.
    cmap : str, optional
        The colormap to use.
    interpolation : str, optional
        Interpolation method used by `imshow`.
    colorbar : bool, optional
        Whether to add a colorbar.
    cbar_label : str, optional
        Label for the colorbar.
    title : str, optional
        Title of the plot.
    xlabel : str, optional
        Label for the x-axis.
    ylabel : str, optional
        Label for the y-axis.
    **kwargs :
        Additional keyword arguments passed to `numpy.histogram2d`.

    Returns
    -------
    H : numpy.ndarray
        The 2D histogram array.
    im : matplotlib.image.AxesImage
        The image object returned by `imshow`.

    Examples
    --------
    >>> data = np.random.randn(1000, 3)
    >>> H, im = histogram2d(data, bins=50, log=True, title="2D Histogram")
    >>> plt.show()

    >>> weights = np.random.rand(1000)
    >>> H, im = histogram2d(data, bins=50, to_weight=weights, cbar_label="Average Weight")
    >>> plt.show()
    """
    if ax is None:
        ax = plt.gca()

    if not isinstance(plane, str) or len(plane)>2:
        raise(ValueError("`plane` argument must indicate in which plane perform the scatter plot (e.g. 'xy', 'Yx', 'Xz', etc...)."))
    
    index1 = coordinate_indices[plane[0]]
    index2 = coordinate_indices[plane[1]]

    if isinstance(vectors3d, cp.ndarray):
        vectors3d = vectors3d.get()
    
    x = vectors3d[:,index1] / scale
    y = vectors3d[:,index2] / scale

    where = (~np.isnan(x)) & (~np.isnan(y))
    x = x[where]
    y = y[where]

    if range is None and dx is not None:
        xc = np.nanmean(x)
        yc = np.nanmean(y)
        range = [[xc - dx, xc + dx], [yc - dx, yc + dx]]

    if to_weight is None:
        H, xedges, yedges = np.histogram2d(x, y, bins=bins, range=range, **kwargs)
    else:
        to_weight = to_weight[where]
        H, xedges, yedges = np.histogram2d(x, y, bins=bins, range=range, weights=to_weight, **kwargs)
        n, xedges, yedges = np.histogram2d(x, y, bins=bins, range=range, **kwargs)
        with np.errstate(divide='ignore', invalid='ignore'):
            H = np.nan_to_num(H / n)

    if update_this is not None:
        H += update_this

    if log:
        H = np.log1p(H)

    if norm is None:
        if log:
            norm = colors.LogNorm(vmin=vmin or np.nanmin(H[H > 0]), vmax=vmax or np.nanmax(H))
        else:
            norm = colors.Normalize(vmin=vmin, vmax=vmax)
        vmin = None
        vmax = None
    
    im = ax.imshow(H.T, origin='lower', extent=[xedges[0], xedges[-1], yedges[0], yedges[-1]], norm=norm, cmap=cmap, vmin=vmin, vmax=vmax, interpolation=interpolation)
    ax.set_aspect(1)

    if colorbar:
        cbar = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)
        if cbar_label:
            cbar.set_label(cbar_label)

    if title:
        ax.set_title(title)
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    return H, im


import os
import yaml
from pathlib import Path, PosixPath, WindowsPath

class PathLoader(yaml.SafeLoader):
    """
    Custom YAML loader that resolves file paths relative to the configuration file location.

    This loader enables the use of the ``!path`` tag in YAML files. When this tag is encountered,
    the loader automatically resolves the specified path relative to the directory containing
    the YAML file, returning a pathlib.Path object.
    """

    def __init__(self, stream):
        # Determine the folder of the file being loaded
        if hasattr(stream, 'name') and stream.name != '<stdout>':
            self._root = os.path.split(stream.name)[0]
        else:
            self._root = os.getcwd()
        super().__init__(stream)

    def construct_path(self, node):
        value = self.construct_scalar(node)
        # Return a Path object so the Dumper can catch it
        full_path = os.path.normpath(os.path.join(self._root, value))
        return Path(full_path)

class PathDumper(yaml.SafeDumper):
    """
    Custom YAML dumper that formats Path objects with the ``!path`` tag.

    If the Path object is absolute, this dumper will automatically attempt to 
    make it relative to the directory of the file being written to.
    """
    def __init__(self, stream, **kwargs):
        self._root = os.path.dirname(stream.name) or '.' if hasattr(stream, 'name') else '.'
        super().__init__(stream, **kwargs)

    def represent_path(self, data):
        try:
            rel_path = os.path.relpath(str(data), self._root)
        except ValueError:
            rel_path = str(data)
        return self.represent_scalar('!path', rel_path.replace(os.sep, '/'))

# Register the constructor for the Loader
PathLoader.add_constructor('!path', PathLoader.construct_path)

# Register the representers for the Dumper
PathDumper.add_representer(Path, PathDumper.represent_path)
PathDumper.add_representer(PosixPath, PathDumper.represent_path)
PathDumper.add_representer(WindowsPath, PathDumper.represent_path)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import cupy as cp
from abc import ABC, abstractmethod

from ..._cpu_transforms import local_to_pointing_rotation

class AbstractPhotonDirections(ABC):
    """
    Abstract base class for a photon source direction generator.

    This class defines the interface for generating photon directions,
    specified by altitude and azimuth angles.

    Attributes
    ----------
    altitude : float
        Altitude angle of the source (in degrees).
    azimuth : float
        Azimuth angle of the source (in degrees).
    """
    def __init__(self, altitude: float, azimuth: float, dtype: type = cp.float32, seed: int | None = None):
        self._altitude = altitude
        self._azimuth = azimuth
        self.dtype = dtype
        self.seed = seed
        self._rot_to_local = cp.asarray(local_to_pointing_rotation(self._altitude, self._azimuth).T, dtype=self.dtype)
    
    @property
    def seed(self):
        """Seed to initialize the CuPy random number generator (cuRAND Philox4x3210)."""
        return self._seed
    
    @seed.setter
    def seed(self, a_seed):
        self.rng = cp.random.default_rng(cp.random.Philox4x3210(a_seed))
        self._seed = a_seed

    @property
    def altitude(self):
        """Source altitude (deg)."""
        return self._altitude

    @altitude.setter
    def altitude(self, alt):
        self._altitude = alt
        self._rot_to_local = cp.asarray(local_to_pointing_rotation(self._altitude, self._azimuth).T, dtype=self.dtype)

    @property
    def azimuth(self):
        """Source azimuth (deg)."""
        return self._azimuth

    @azimuth.setter
    def azimuth(self, az):
        self._azimuth = az
        self._rot_to_local = cp.asarray(local_to_pointing_rotation(self._altitude, self._azimuth).T, dtype=self.dtype)
    
    @abstractmethod
    def generate(self, n: int) -> cp.ndarray:
        """
        Generates photon direction vectors.

        Parameters
        ----------
        n : int
            Number of photon directions to generate.

        Returns
        -------
        directions : cp.ndarray
            A CuPy array of shape (n, 3) where each row represents a
            photon direction vector (unit vector).

        Raises
        ------
        ValueError
            If n is not a positive integer.
        """
        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")
        pass

class PointLike(AbstractPhotonDirections):
    """
    Generates a set of identical direction vector from a source specified in 
    horizontal coordinates.

    This class creates an array of `n` identical unit vectors that represent
    the direction of a source specified by its altitude and azimuth angles.

    Parameters
    ----------
    altitude : float
        Altitude angle of the source in degrees.
    azimuth : float
        Azimuthal angle of the source in degrees (0 degrees corresponds to the
        positive x-axis, and the angle increases counterclockwise).
    dtype : data-type, optional
        Desired output data-type. Defaults to `cp.float32`.
    """

    def __init__(self, altitude: float, azimuth: float, dtype: type = cp.float32):
        super().__init__(altitude, azimuth, dtype=dtype)

    def generate(self, n: int) -> cp.ndarray:
        """
        Generates the set of identical direction vectors.

        Parameters
        ----------
        n : int
            Number of identical direction vectors to generate.

        Returns
        -------
        vs : ndarray, shape (n, 3)
            An array of `n` identical unit vectors representing the direction of
            the source in a right-handed Cartesian coordinate system.

        Raises
        ------
        ValueError
            If `n` is not a positive integer.
        """

        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")

        vs: cp.ndarray = cp.empty((n, 3), dtype=self.dtype)
        vs[:, :2] = 0.
        vs[:, 2] = 1.

        vs = vs @ self._rot_to_local.T
        vs[:, :] *= -1.

        return vs

class UniformBeam(AbstractPhotonDirections):
    """
    Generates uniformly distributed directions within a cone around a central direction.

    This class provides a way to generate `n` unit vectors that are uniformly
    distributed within a cone defined by a central direction (`altitude`, `azimuth`)
    and an aperture angle. The distribution is uniform in the sense that the
    probability of finding a vector within a given solid angle is proportional
    to that solid angle.

    Parameters
    ----------
    altitude : float
        Altitude angle of the central direction of the cone (in degrees).
    azimuth : float
        Azimuth angle of the central direction of the cone (in degrees).
    aperture : float
        Angular half-size (radius) of the cone (in degrees).
    dtype : data-type, optional
        Desired data type of the output array, by default cupy.float32
    seed : int | None, optional
        A seed to initialize the CuPy random number generator (cuRAND Philox4x3210). If None, the
        generator is seeded using unpredictable data from the operating system.
    """

    def __init__(
            self,
            altitude: float,
            azimuth: float,
            half_aperture: float,
            dtype: type = cp.float32,
            seed: int | None = None
        ):
        super().__init__(altitude, azimuth, dtype=dtype, seed=seed)
        self.half_aperture = half_aperture

    def generate(self, n: int) -> cp.ndarray:
        """
        Generates `n` uniformly distributed direction vectors within the cone.

        Parameters
        ----------
        n : int
            Number of direction vectors to generate.

        Returns
        -------
        directions : ndarray, shape (n, 3)
            A CuPy array of shape (n, 3) containing the generated direction vectors.
            Each row represents a unit vector (x, y, z).

        Raises
        ------
        ValueError
            If `n` is not a positive integer.
        """
        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")

        # Ensure uniform distribution over the sphere's surface within the cone
        phis = 2 * cp.pi * self.rng.random(size=n)
        aperture_rad = self.half_aperture / 180. * cp.pi
        cos_theta_min = cp.cos(aperture_rad)
        cos_thetas = self.rng.uniform(cos_theta_min, 1., n)
        thetas = cp.arccos(cos_thetas)

        # Allocate array for direction vectors
        vs = cp.empty((n, 3), dtype=self.dtype)

        sith = cp.sin(thetas)

        # Convert spherical coordinates (altitude, azimuth) to Cartesian coordinates (x, y, z)
        vs[:, 0] = cp.cos(phis) * sith  # x-component
        vs[:, 1] = cp.sin(phis) * sith  # y-component
        vs[:, 2] = cos_thetas #cp.cos(thetas)  # z-component

        # Apply the rotation to the direction vectors
        vs = vs @ self._rot_to_local.T

        # Flip the z-axis
        vs[:, :] *= -1.

        return vs

class GaussianBeam(AbstractPhotonDirections):
    """
    Generates direction vectors with a Gaussian distribution around a central direction.

    This class generates `n` direction vectors sampled from a Gaussian
    distribution centered around the direction specified by `altitude` and `azimuth`.
    The spread of the distribution is controlled by `dispersion`.

    Parameters
    ----------
    altitude : float
        Altitude angle of the beam center in degrees.
    azimuth : float
        Azimuthal angle of the beam center in degrees.
    dispersion : float
        Standard deviation (sigma) of the Gaussian distribution in degrees.
    seed : int | None, optional
        A seed to initialize the CuPy random number generator (cuRAND Philox4x3210). If None, the
        generator is seeded using unpredictable data from the operating system.
    dtype : data-type, optional
        Desired output data-type. Default is `cp.float32`.

    """
    def __init__(self, altitude: float, azimuth: float, dispersion: float, seed: int | None = None, dtype: type = cp.float32):
        super().__init__(altitude, azimuth, dtype=dtype, seed=seed)
        self.dispersion = dispersion
    
    def generate(self, n: int) -> cp.ndarray:
        """
        Generates the direction vectors sampled from the Gaussian beam.

        Parameters
        ----------
        n : int
            Number of direction vectors to generate. Must be a positive integer.

        Returns
        -------
        vs : ndarray, shape (n, 3)
            An array of `n` unit vectors representing the directions sampled from
            the Gaussian beam in a right-handed Cartesian coordinate system.
        
        Raises
        ------
        ValueError
            If `n` is not a positive integer.
        """

        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")

        logu = cp.log(self.rng.random(size=n))
        thetas = self.dispersion / 180 * cp.sqrt(-2. * logu) * cp.pi
        phis = self.rng.random(size=n) * 2. * cp.pi

        sin_theta = cp.sin(thetas)
        cos_theta = cp.cos(thetas)
        sin_phi = cp.sin(phis)
        cos_phi = cp.cos(phis)

        # Allocate array for direction vectors
        vs = cp.empty((n, 3), dtype=self.dtype)
        vs[:, 0] = sin_theta * cos_phi
        vs[:, 1] = -sin_theta * sin_phi
        vs[:, 2] = cos_theta

        # Apply the rotation to the direction vectors
        vs = vs @ self._rot_to_local.T

        # Flip the z-axis
        vs[:, :] *= -1.

        return vs


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import importlib.resources
import functools

from ..utils._is_notebook import is_notebook

class MplInteractiveContext:
    """Re-enable interactive mode.
    See issue:
        matplotlib.style.context() stops plotting inline in Jupyter
        https://github.com/matplotlib/matplotlib/issues/26716
    """
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        import matplotlib.pyplot as plt
        if is_notebook():
            plt.ion()

def iactsim_style(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import matplotlib.pyplot as plt
        with MplInteractiveContext():
            with importlib.resources.path('iactsim', 'iactsim.mplstyle') as style_path:
                with plt.style.context(style_path):
                    result = func(*args, **kwargs)
        
        return result
    return wrapper


import pytest
import cupy as cp
import numpy as np
from iactsim.electronics.signals.trigger_logic import topological_camera_trigger

CONSTANTS = {
    "WINDOW_SIZE": 50,
    "MODULE_DIM": 8,
    "N_MODULES": 5,
    "N_THREADS": 128,
    "SEED": 12345
}

SHAPES = {
    'line_v_3': [(0,0), (1,0), (2,0)],             
    'line_v_4': [(0,0), (1,0), (2,0), (3,0)],      
    'square_2x2': [(0,0), (0,1), (1,0), (1,1)],    
    'L_shape_4': [(0,0), (1,0), (2,0), (2,1)],     
    'T_shape_4': [(0,0), (1,0), (2,0), (1,1)],     
    'cross_5':   [(1,0), (0,1), (1,1), (2,1), (1,2)], 
    'disjoint_4': [(0,0), (0,1), (5,5), (5,6)]     
}

def xy_to_id(x, y, dim=8):
    """Convert (x, y) to flat pixel index."""
    if 0 <= x < dim and 0 <= y < dim:
        return x * dim + y
    return None

@pytest.fixture
def gpu_memory():
    """
    Allocates GPU memory. Returns a dict containing the arrays.
    """
    n_pixels = CONSTANTS["MODULE_DIM"] ** 2
    total_size = CONSTANTS["N_MODULES"] * n_pixels * CONSTANTS["WINDOW_SIZE"]
    
    # Inputs
    disc_signals = cp.zeros(total_size, dtype=cp.float32)
    
    # Pixel stack buffer
    ints_per_thread = n_pixels + (n_pixels // 32) + 2
    stack_size = CONSTANTS["N_MODULES"] * CONSTANTS["N_THREADS"] * ints_per_thread
    stack_buffer = cp.empty(stack_size, dtype=cp.int32)
    
    # Outputs
    mod_trig_sig = cp.zeros(CONSTANTS["N_MODULES"] * CONSTANTS["WINDOW_SIZE"], dtype=cp.float32)
    dummy_1 = cp.full((2,), -1, dtype=cp.int32)
    
    return {
        "disc_signals": disc_signals,
        "stack_buffer": stack_buffer,
        "mod_trig_sig": mod_trig_sig,
        "dummy_1": dummy_1,
    }

def run_trigger_kernel(gpu_mem, n_contiguous):
    """Launch the kernel with current memory and config."""
    topological_camera_trigger(
        (CONSTANTS["N_MODULES"],), (CONSTANTS["N_THREADS"],),
        (
            gpu_mem["disc_signals"], 
            n_contiguous, 
            gpu_mem["mod_trig_sig"],
            gpu_mem["dummy_1"], 
            gpu_mem["stack_buffer"],
            CONSTANTS["WINDOW_SIZE"], 
            CONSTANTS["MODULE_DIM"], 
            CONSTANTS["N_MODULES"], 
            CONSTANTS["SEED"]
        )
    )

def inject_shape(gpu_mem, mod_id, time, shape_coords, offset=(0,0)):
    """Injects a shape into the signals array."""
    dim = CONSTANTS["MODULE_DIM"]
    win = CONSTANTS["WINDOW_SIZE"]
    n_pix = dim * dim
    base_idx = mod_id * n_pix * win

    active_pixels = []
    for (sx, sy) in shape_coords:
        px, py = sx + offset[0], sy + offset[1]
        pid = xy_to_id(px, py, dim)
        if pid is not None:
            idx = base_idx + win * pid + time
            gpu_mem["disc_signals"][idx] = 1.0
            active_pixels.append(pid)
    return active_pixels

@pytest.mark.parametrize("shape_name, n_contiguous, should_trigger", [
    # Basic threshold checks
    ('line_v_3',   3, True),   # 3-pixel line, thresh 3 -> Pass
    ('line_v_3',   4, False),  # 3-pixel line, thresh 4 -> Fail
    ('line_v_4',   4, True),   # 4-pixel line, thresh 4 -> Pass
    ('square_2x2', 4, True),   # 4-pixel square, thresh 4 -> Pass
    ('cross_5',    4, True),   # 5-pixel cross, thresh 4 -> Pass
    ('cross_5',    6, False),  # 5-pixel cross, thresh 6 -> Fail
    
    # Topology logic checks
    ('disjoint_4', 3, False),  # 4 pixels total, but separated
    ('L_shape_4',  4, True),   # L-shape is contiguous
    ('T_shape_4',  4, True),   # T-shape is contiguous
])
def test_shapes_and_thresholds(gpu_memory, shape_name, n_contiguous, should_trigger):
    """
    Verifies that various shapes trigger (or don't) based on pixel count and contiguous logic.
    """
    mod_id = 0
    time = 20
    
    inject_shape(gpu_memory, mod_id, time, SHAPES[shape_name])
    
    run_trigger_kernel(gpu_memory, n_contiguous)
    
    # Check result
    res_idx = mod_id * CONSTANTS["WINDOW_SIZE"] + time
    triggered = gpu_memory["mod_trig_sig"][res_idx] > 0
    
    assert triggered == should_trigger, \
        f"Shape {shape_name} with n={n_contiguous} failed. Expected {should_trigger}, got {triggered}"


@pytest.mark.parametrize("offset_x, offset_y", [
    (0, 0),  # Top-Left
    (6, 6),  # Bottom-Right (for 2x2 shape in 8x8 grid)
    (0, 6),  # Edge
    (6, 0),  # Edge
    (3, 3)   # Center
])
def test_sliding_window_robustness(gpu_memory, offset_x, offset_y):
    """
    Verifies a 2x2 Square triggers correctly regardless of position in the module.
    """
    mod_id = 1
    time = 15
    n_contiguous = 4
    
    # Inject Square
    inject_shape(gpu_memory, mod_id, time, SHAPES['square_2x2'], offset=(offset_x, offset_y))
    
    run_trigger_kernel(gpu_memory, n_contiguous)
    
    res_idx = mod_id * CONSTANTS["WINDOW_SIZE"] + time
    triggered = gpu_memory["mod_trig_sig"][res_idx] > 0
    
    assert triggered, f"Failed to trigger at offset {offset_x},{offset_y}"


def test_inter_module_isolation(gpu_memory):
    """
    Verifies that pixels split between two modules do not sum up to trigger.
    """
    time = 25
    n_contiguous = 4
    
    # Module 0 Right edge
    inject_shape(gpu_memory, 0, time, [(0,0), (1,0)], offset=(6, 7))

    p_mod0 = [(0, 7), (1, 7)]
    inject_shape(gpu_memory, 0, time, p_mod0)

    # Module 1: Left Edge (2 pixels)
    p_mod1 = [(0, 0), (1, 0)]
    inject_shape(gpu_memory, 1, time, p_mod1)

    run_trigger_kernel(gpu_memory, n_contiguous)
    
    # Assertions
    idx_0 = 0 * CONSTANTS["WINDOW_SIZE"] + time
    idx_1 = 1 * CONSTANTS["WINDOW_SIZE"] + time
    
    assert gpu_memory["mod_trig_sig"][idx_0] == 0, "Module 0 triggered incorrectly on partial shape"
    assert gpu_memory["mod_trig_sig"][idx_1] == 0, "Module 1 triggered incorrectly on partial shape"


def test_time_isolation(gpu_memory):
    """
    Verifies that a shape spread across two different time slices does not trigger.
    """
    mod_id = 2
    n_contiguous = 4
    
    # Inject 2 pixels at T=10
    inject_shape(gpu_memory, mod_id, 10, [(0,0), (0,1)])
    
    # Inject 2 pixels at T=11 (physically adjacent to previous ones, but different time)
    inject_shape(gpu_memory, mod_id, 11, [(1,0), (1,1)])
    
    run_trigger_kernel(gpu_memory, n_contiguous)
    
    idx_10 = mod_id * CONSTANTS["WINDOW_SIZE"] + 10
    idx_11 = mod_id * CONSTANTS["WINDOW_SIZE"] + 11
    
    assert gpu_memory["mod_trig_sig"][idx_10] == 0, "Triggered at T=10 incorrectly"
    assert gpu_memory["mod_trig_sig"][idx_11] == 0, "Triggered at T=11 incorrectly"


import pytest
import cupy as cp
import iactsim
import numpy as np

@pytest.fixture(scope="module")
def telescope_setup():
    """Initializes the telescope and GPU context."""
    astri = iactsim.models.astri.AstriTelescope()

    astri.optical_system.build_m1_segments()

    astri.optical_system.build_masts()

    astri.cuda_init()

    return astri

def test_ray_tracing_performance_astri(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.set_target(distance=11000)

    def setup_round_on_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_on_axis, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_astri_off_axis(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)
    def setup_round_off_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_off_axis, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_astri_pdms(benchmark, telescope_setup):
    telescope = telescope_setup
    telescope.optical_system.build_camera_modules()
    telescope.cuda_init()
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_astri_pdms_window(benchmark, telescope_setup):
    telescope = telescope_setup
    telescope.optical_system.build_camera_modules()
    telescope.optical_system.set_camera_window_parameters(
        240, 8.61, 1.5, 1, 17,
        layer_1_as_cylinder = True,
        layer_2_as_cylinder = True,
        layer_3_as_cylinder = True
    )
    telescope.cuda_init()
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_astri_pdms_window_props(benchmark, telescope_setup):
    telescope = telescope_setup

    telescope.optical_system.build_camera_modules()

    telescope.optical_system.set_camera_window_parameters(
        240, 8.61, 1.5, 1, 17,
        layer_1_as_cylinder = True,
        layer_2_as_cylinder = True,
        layer_3_as_cylinder = True
    )

    gen = iactsim.optics.FresnelSurfacePropertiesGenerator()
    wls = np.arange(200,1001,1.)
    angls = np.arange(0,91,1.)
    gen.generate(telescope.optical_system['WindowLayer1'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer2'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer3'], wls, angls,inplace=True)
    telescope.cuda_init()

    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_astri_pdms_props_window_props(benchmark, telescope_setup):
    # SiPM optical properties
    sipm = iactsim.optics.SipmStackSimulator(
        layers=["SiO2", "ZrO2"],
        thicknesses=[3250, 35],
        medium="Air",
        type_layers=['c', 'c'],
        depletion_layer_depth=3.,
        depletion_layer_width=2600,
        fill_factor=0.75
    )
    sipm.set_simulation_params(
        wl_range=np.arange(200,1000,1),
        angle_range=np.arange(0,91,1)
    )
    sipm.run('u')
    sipm.constrain_pde(462, 0.582517*0.97)
    sipm_surface_prop = sipm.get_surfcace_properties_obj()

    telescope = telescope_setup

    telescope.optical_system.build_camera_modules()

    telescope.optical_system.set_camera_window_parameters(
        240, 8.61, 1.5, 1, 17,
        layer_1_as_cylinder = True,
        layer_2_as_cylinder = True,
        layer_3_as_cylinder = True
    )

    gen = iactsim.optics.FresnelSurfacePropertiesGenerator()
    wls = np.arange(200,1001,1.)
    angls = np.arange(0,91,1.)
    gen.generate(telescope.optical_system['WindowLayer1'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer2'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer3'], wls, angls,inplace=True)

    for s in telescope.optical_system:
        if s.name.startswith('PDM'):
            s.properties = sipm_surface_prop

    telescope.cuda_init()

    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)
    
    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1, warmup_rounds=5)